/*

*/

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.Random;

/**
   Implement this task graph using "message passing".

       task1    task2    task3
            \     |     /
             \    |    /
              \   |   /
                task4
              /   |   \
             /    |    \
            /     |     \
       task5    task6    task7
*/
public class TaskGraph
{
   public static void main(String[] args) throws InterruptedException
   {
      final Task task1 = new Task(1, 1);
      final Task task2 = new Task(2, 2);
      final Task task3 = new Task(3, 3);
      final Task task4 = new Task(4, 2);
      final Task task5 = new Task(5, 1);
      final Task task6 = new Task(6, 2);
      final Task task7 = new Task(7, 3);

      final var future1 = new CompletableFuture<Integer>();
      final var future2 = new CompletableFuture<Integer>();
      final var future3 = new CompletableFuture<Integer>();
      final var future4 = new CompletableFuture<Integer>();
      final var future5 = new CompletableFuture<Integer>();
      final var future6 = new CompletableFuture<Integer>();

      final Thread t1 = new Thread(
                               ()->{ task1.run();
                                     future1.complete(0); // send a message
                                   }
                                  );

      final Thread t2 = new Thread(
                               ()->{ task2.run();
                                     future2.complete(0); // send a message
                                   }
                                  );

      final Thread t3 = new Thread(
                               ()->{ task3.run();
                                     future3.complete(0); // send a message
                                   }
                                  );

      final Thread t4 = new Thread(
                               ()->{try{ future1.get(); // receive a message (blocks)
                                         future2.get(); // receive a message (blocks)
                                         future3.get(); // receive a message (blocks)
                                         task4.run();
                                         future4.complete(0); // send a message
                                         future5.complete(0); // send a message
                                         future6.complete(0); // send a message
                                       }catch(ExecutionException
                                             |InterruptedException e){}}
                                  );

      final Thread t5 = new Thread(
                               ()->{try{ future4.get(); // receive a message (blocks)
                                         task5.run();
                                       }catch(ExecutionException
                                             |InterruptedException e){}}
                                  );

      final Thread t6 = new Thread(
                               ()->{try{ future5.get(); // receive a message (blocks)
                                         task6.run();
                                       }catch(ExecutionException
                                             |InterruptedException e){}}
                                  );

      final Thread t7 = new Thread(
                               ()->{try{ future6.get(); // receive a message (blocks)
                                         task7.run();
                                       }catch(ExecutionException
                                             |InterruptedException e){}}
                                  );
      t1.start();
      t2.start();
      t3.start();
      t4.start();
      t5.start();
      t6.start();
      t7.start();
   }


   static class Task implements Runnable
   {
      private final int id;
      private final int colunmNumber;
      private final Random random;

      Task(int id, int columnNumber)
      {
         this.id = id;
         this.colunmNumber = columnNumber;
         this.random = new Random();
      }

      @Override
      public void run()
      {
         // Tab over to the column used by this task.
         final String indent = "\t".repeat(colunmNumber); // Java 11.

         // How long this task should "work".
         int computeTime = 2 + random.nextInt(8); // Between 2 and 10 seconds.

         // Print a message in the appropriate column for this task.
         // The message shows this task's id and the number of seconds
         // left for this task to execute.
         System.out.println(indent + "t" + id + "," + computeTime);
         while (computeTime > 0)
         {
            try{ TimeUnit.SECONDS.sleep(1); }catch(InterruptedException e){}
            --computeTime;
            System.out.println(indent + "t" + id + "," + computeTime);
         }
      }
   }
}
